module read_hdfeos_module

use constants_Module
use hdfeos5_Module
use datastruct_module
use datetime_module

contains
!  call read_omi_tropno2(filename,temp_no2data,dumA,dumB,nxtrack,nimages,npressurelevels,.false.,status)
  
subroutine read_OMI_tropNO2(filename,datastruct,a_lev,b_lev,nxtrack,ntimes,npressurelevels,verbose,retstatus)
  implicit none

  real,dimension(:,:),allocatable           :: SolarZenithAngle,SolarAzimuthAngle,longitude,latitude,SurfaceAlbedo
  real,dimension(:,:),allocatable           :: ViewingZenithAngle,ViewingAzimuthAngle,TM4SurfacePressure
  real,dimension(:),allocatable             :: SpacecraftAltitude,SpacecraftLatitude,SpacecraftLongitude,floatbuffer
  real,dimension(:,:),allocatable           :: GhostColumn,AirMassFactor,AirMassFactorTropospheric,AirMassFactorGeometric
  real,dimension(:,:),allocatable           :: VCDTropErrorUsingAvKernel,VCDErrorUsingAvKernel,CloudRadianceFraction
  real,dimension(:,:),allocatable           :: TroposphericVerticalColumnError  
  real,dimension(:,:),allocatable           :: TotalVerticalColumn,TotalVerticalColumnError,TroposphericVerticalColumn
  real,dimension(:,:),allocatable           :: AssStratSlantColumn,AssStratVerticalColumn
  real,dimension(:,:),allocatable           :: AssStratVerticalColumnError,cloudfraction,cloudpressure,cloudfractionstd,cloudpressurestd
  real,dimension(:,:),allocatable           :: SlantColumnAmountNO2,SlantColumnAmountNO2Std,TroposphericVerticalColumnModel
  real,dimension(:,:,:),allocatable         :: LatitudeCornerpoints,LongitudeCornerpoints,averagingkernel
  real(flt64),   dimension(:),  allocatable :: TAITime
  real,dimension(MAX_presslevels)           :: a_lev,b_lev
  integer(int32),dimension(:),allocatable   :: htime,hdate
  integer(int16),dimension(:,:),allocatable :: GroundPixelQualityFlags,int16_buffer,TerrainHeight,TM4TerrainHeight
  integer(int16),dimension(:,:,:),allocatable :: int16_buffer3d
  integer(int32),dimension(:,:),allocatable :: int32_buffer
  integer(int_8), dimension(:),allocatable  :: MeasurementQualityFlags,InstrumentConfigurationId
  integer(int_8),dimension(:,:),allocatable :: TM4TropoPauseLevel,TroposphericColumnFlag,int8_buffer
  integer(int32)                            :: itau_offset_TM_OMI
  integer(int32)                            :: numbertype

  integer,intent(out)                                   :: retstatus
  integer                                               :: fid,status,swid
  integer                                               :: rank,dims(3),i
  integer                                               :: nxtrack,ntimes,npressurelevels
  integer(hdf_kind_type)                                :: start(3),stride(3),edge(3)
  real                                                  :: scalefactor
  logical                                               :: fexist
  logical,intent(in)                                    :: verbose
  character(len=STD_STRING)                             :: dimlist, maxdimlist
  character(len=STD_STRING),intent(in)                  :: filename
  character(len=SWATHNAME_LENGTH)                       :: swathname,fieldname
  character(len=5000)                                   :: dfieldlist,gfieldlist
  integer                                               :: ndflds,ngflds,pos,dumtime(6)
  integer,dimension(100)                                :: ntype,nrank
  type(Tdatastruct),dimension(max_nxtrack,max_ntimes)   :: datastruct

  retstatus = 1 ; fid = -1 ; swid = -1
  write(swathname,'(a)') '                                                          '
  write(swathname,'(a)') 'DominoNO2'
  inquire (file=filename,exist=fexist)
  
  if (fexist == .false.) then
    write(*,'(3a)') 'read_OMI_tropNO2: ERROR file ',trim(filename),' does not exist'
    return
  end if

  ! open the file with read-only access
  fid = he5_swopen( filename, HE5F_ACC_RDONLY )         
  if (fid <= 0 ) then
    write(*,'(2a)') 'read_OMI_tropNO2: ERROR cannot open file ',trim(filename)
    return
  end if
  if (verbose) write(*,'(3a)') 'File ',trim(filename),' opened'
  
  ! attach to the swath
  swid = he5_swattach( fid, swathname )  
  if (swid <= 0 ) then
    write(*,'(2a)') 'read_OMI_tropNO2: ERROR cannot attach swath ',trim(swathname)
    status = he5_swclose(fid)
    return
  end if
  if (verbose) write(*,'(2a)') 'Attached swath ',trim(swathname)

  ndflds = he5_swinqdflds(swid,dfieldlist,nrank,ntype)
  ngflds = he5_swinqgflds(swid,gfieldlist,nrank,ntype)
  if (verbose) write(*,'("nr of data fields: ",i3," nr of geo fields: ",i3)') ndflds,ngflds
  
  pos = index(trim(gfieldlist),'Latitude')
  if (pos < 1) then
    write(*,'(a)') 'The file is invalid, it doesnt contain the Latitude field'
    return
  end if

  ! get dimensions
  fieldname = 'AveragingKernel'
  status = he5_swfldinfo(swid, fieldname, rank, dims,numbertype, dimlist, maxdimlist)  
  if (status /= 0 ) then
    write(*,'(a)') 'read_OMI_tropNO2: ERROR retrieving fieldinfo'
    status = he5_swdetach(swid)
    status = he5_swclose(fid)
    return
  end if

  nxtrack = dims(1)
  ntimes  = dims(2)
  npressurelevels = dims(3)

  if (verbose) write(*,'("The orbit holds ",i3," x ",i4," pixels Npressure =",i)') dims(1),dims(2),dims(3)

  if ( (nxtrack > MAX_NXTRACK) .OR. (nxtrack <= 0) ) then
    write(*,'(a,i)') 'read_OMI_tropNO2: ERROR invalid value of nxtrack ',nxtrack
    status = he5_swdetach(swid)
    status = he5_swclose(fid)
    return
  end if

  if ( (ntimes > MAX_NTIMES) .OR. (ntimes <= 0)) then
    write(*,'(a,i)') 'read_OMI_tropNO2: ERROR invalid value of ntimes ',ntimes
    status = he5_swdetach(swid)
    status = he5_swclose(fid)
    return
  end if

  if ( (npressurelevels > MAX_presslevels) .OR. (npressurelevels <= 0)) then
    write(*,'(a,i)') 'read_OMI_tropNO2: ERROR invalid value of npressurelevels ',npressurelevels
    status = he5_swdetach(swid)
    status = he5_swclose(fid)
    return
  end if

  allocate(latitude(nxtrack,ntimes))
  allocate(longitude(nxtrack,ntimes))
  allocate(SolarZenithAngle(nxtrack,ntimes))
  allocate(SolarAzimuthAngle(nxtrack,ntimes))
  allocate(ViewingZenithAngle(nxtrack,ntimes))
  allocate(ViewingAzimuthAngle(nxtrack,ntimes))
  allocate(cloudfraction(nxtrack,ntimes))
  allocate(cloudpressure(nxtrack,ntimes))
  allocate(terrainheight(nxtrack,ntimes))
  allocate(TM4SurfacePressure(nxtrack,ntimes))
  allocate(TAItime(ntimes))
  allocate(htime(ntimes))
  allocate(hdate(ntimes))  
  allocate(SpacecraftAltitude(ntimes))
  allocate(SpacecraftLatitude(ntimes))
  allocate(SpacecraftLongitude(ntimes))
  allocate(GhostColumn(nxtrack,ntimes))
  allocate(AirMassFactor(nxtrack,ntimes))
  allocate(AirMassFactorTropospheric(nxtrack,ntimes))
  allocate(AirMassFactorGeometric(nxtrack,ntimes))
  allocate(SurfaceAlbedo(nxtrack,ntimes))
  allocate(VCDTropErrorUsingAvKernel(nxtrack,ntimes))
  allocate(VCDErrorUsingAvKernel(nxtrack,ntimes))
  allocate(CloudRadianceFraction(nxtrack,ntimes))
  allocate(TroposphericVerticalColumnError(nxtrack,ntimes))
  allocate(TotalVerticalColumn(nxtrack,ntimes))
  allocate(TotalVerticalColumnError(nxtrack,ntimes))
  allocate(TroposphericVerticalColumn(nxtrack,ntimes))
  allocate(AssStratSlantColumn(nxtrack,ntimes))
  allocate(AssStratVerticalColumn(nxtrack,ntimes))
  allocate(AssStratVerticalColumnError(nxtrack,ntimes)) ; AssStratVerticalColumnError=0.25
  allocate(SlantColumnAmountNO2(nxtrack,ntimes))
  allocate(SlantColumnAmountNO2Std(nxtrack,ntimes))
  allocate(TroposphericVerticalColumnModel(nxtrack,ntimes))
  allocate(LatitudeCornerpoints(nxtrack,ntimes,NCORNERPOINTS))
  allocate(LongitudeCornerpoints(nxtrack,ntimes,NCORNERPOINTS))
  allocate(TM4TropoPauseLevel(nxtrack,ntimes)) 
  allocate(TM4TerrainHeight(nxtrack,ntimes))   
  allocate(TroposphericColumnFlag(nxtrack,ntimes))
  allocate(InstrumentConfigurationId(ntimes))
  allocate(MeasurementQualityFlags(ntimes))
  allocate(GroundPixelQualityFlags(nxtrack,ntimes))
  allocate(CloudPressureStd(nxtrack,ntimes))
  allocate(CloudFractionStd(nxtrack,ntimes))
  allocate(int16_buffer(nxtrack,ntimes))
  allocate(int32_buffer(nxtrack,ntimes))
  allocate(int8_buffer(nxtrack,ntimes))  
  allocate(averagingkernel(nxtrack,ntimes,npressurelevels))
  allocate(int16_buffer3d(nxtrack,ntimes,npressurelevels))
  allocate(floatbuffer(npressurelevels))

  
  ! 1-D fields
  start(:)  = 0
  stride(:) = 1
  edge(1)   = ntimes

  taitime = -1. ; htime = -1 ; hdate = -1
  
  retstatus = 10 ; fieldname =  'Time'
  if ( index(trim(gfieldlist),trim(fieldname)) > 0 ) then 
    status = he5_swrdfld( swid, fieldname, start, stride, edge, TAItime)
    if (status /= 0 ) then 
      write(*,'(2a)') 'ERROR reading ',fieldname ; status = he5_swdetach(swid)
      status = he5_swclose(fid) ; goto 100 
    end if
  end if

  iyear0 = 1950 ; icalendo = 2 ; julday0 = julday(1,1,iyear0)
  call getTMTimeOffset ( itau_offset_TM_OMI )
  do i=1,edge(1)
    if (taitime(i) < 3.e8) cycle
    call tau2date(int(TAITime(i) - TAI_20040101UTC + itau_offset_TM_OMI),dumtime)
    htime(i) = dumtime(4)*10000 + dumtime(5)*100 + dumtime(6)
    hdate(i) = dumtime(1)*10000 + dumtime(2)*100 + dumtime(3)
  end do

  retstatus = retstatus + 1 ; fieldname =  'MeasurementQualityFlags'
  if ( index(trim(dfieldlist),trim(fieldname)) > 0 ) then 
    status = he5_swrdfld( swid, fieldname, start, stride, edge, MeasurementQualityFlags)
    if (status /= 0 ) then 
      write(*,'(2a)') 'ERROR reading ',fieldname ; status = he5_swdetach(swid) 
      status = he5_swclose(fid) ; goto 100 
    end if
  end if
  
  retstatus = retstatus + 1 ; fieldname =  'InstrumentConfigurationId'
  if (index(trim(dfieldlist),trim(fieldname)) > 0) then
    status = he5_swrdfld( swid, fieldname, start, stride, edge, InstrumentConfigurationId)
    if (status /= 0 ) then 
      write(*,'(2a)') 'ERROR reading ',fieldname ; status = he5_swdetach(swid) 
      status = he5_swclose(fid) ; goto 100 
    end if
  end if

  retstatus = retstatus +1 ; fieldname =  'SpacecraftAltitude'
  if (index(trim(gfieldlist),trim(fieldname)) > 0) then
    status = he5_swrdfld( swid, fieldname, start, stride, edge, SpacecraftAltitude)
    if (status /= 0 ) then 
      write(*,'(2a)') 'ERROR reading ',fieldname ; status = he5_swdetach(swid) 
      status = he5_swclose(fid) ; goto 100 
    end if
  end if
 
  retstatus = retstatus +1 ; fieldname =  'SpacecraftLatitude'
  if (index(trim(gfieldlist),trim(fieldname)) > 0) then
    status = he5_swrdfld( swid, fieldname, start, stride, edge, SpacecraftLatitude)
    if (status /= 0 ) then 
      write(*,'(2a)') 'ERROR reading ',fieldname ; status = he5_swdetach(swid) 
      status = he5_swclose(fid) ; goto 100 
    end if
  end if

  retstatus = retstatus +1 ; fieldname =  'SpacecraftLongitude'
  if (index(trim(gfieldlist),trim(fieldname)) > 0) then
    status = he5_swrdfld( swid, fieldname, start, stride, edge,SpacecraftLongitude )
    if (status /= 0 ) then 
      write(*,'(2a)') 'ERROR reading ',fieldname ; status = he5_swdetach(swid) 
      status = he5_swclose(fid) ; goto 100 
    end if
  end if

  ! 2-D fields
  start(:)  = 0 ; stride(:) = 1 
  edge(1)   = nxtrack ; edge(2) = ntimes ; edge(3) = NCORNERPOINTS

  fieldname = 'Latitude' ; retstatus = retstatus + 1
  if (index(trim(gfieldlist),trim(fieldname)) > 0) then
    status = he5_swrdfld( swid, fieldname, start, stride, edge, latitude )
    if (status /= 0 ) then 
      write(*,'(2a)') 'ERROR reading ',fieldname ; status = he5_swdetach(swid) 
      status = he5_swclose(fid) ; goto 100 
    end if
  end if

  fieldname = 'Longitude' ; retstatus = retstatus + 1
  if (index(trim(gfieldlist),trim(fieldname)) > 0) then
    status = he5_swrdfld( swid, fieldname, start, stride, edge, longitude )
    if (status /= 0 ) then 
      write(*,'(2a)') 'ERROR reading ',fieldname ; status = he5_swdetach(swid) 
      status = he5_swclose(fid) ; goto 100 
    end if
  end if

  retstatus = retstatus +1 ; fieldname = 'SolarZenithAngle'
  if (index(trim(gfieldlist),trim(fieldname)) > 0) then
    status = he5_swrdfld( swid, fieldname, start, stride, edge, solarZenithangle )
    if (status /= 0 ) then 
      write(*,'(2a)') 'ERROR reading ',fieldname ; status = he5_swdetach(swid) 
      status = he5_swclose(fid) ; goto 100 
    end if
  end if
  
  retstatus = retstatus +1 ; fieldname =  'SolarAzimuthAngle'
  if (index(trim(gfieldlist),trim(fieldname)) > 0) then
    status = he5_swrdfld( swid, fieldname, start, stride, edge, solarAzimuthangle )
    if (status /= 0 ) then 
      write(*,'(2a)') 'ERROR reading ',fieldname ; status = he5_swdetach(swid) 
      status = he5_swclose(fid) ; goto 100 
    end if
  end if

  retstatus = retstatus +1 ; fieldname =  'ViewingZenithAngle'
  if (index(trim(gfieldlist),trim(fieldname)) > 0) then
    status = he5_swrdfld( swid, fieldname, start, stride, edge, viewingZenithangle )
    if (status /= 0 ) then 
      write(*,'(2a)') 'ERROR reading ',fieldname ; status = he5_swdetach(swid) 
      status = he5_swclose(fid) ; goto 100 
    end if
  end if

  retstatus = retstatus +1 ; fieldname =  'ViewingAzimuthAngle'
  if (index(trim(gfieldlist),trim(fieldname)) > 0) then
    status = he5_swrdfld( swid, fieldname, start, stride, edge, viewingAzimuthangle )
    if (status /= 0 ) then 
      write(*,'(2a)') 'ERROR reading ',fieldname ; status = he5_swdetach(swid) 
      status = he5_swclose(fid) ; goto 100 
    end if
  end if

  retstatus = retstatus +1 ; fieldname =  'CloudFraction'
  if (index(trim(dfieldlist),trim(fieldname)) > 0) then
    status = he5_swrdfld( swid, fieldname, start, stride, edge, int16_buffer )
    if (status /= 0 ) then 
      write(*,'(2a)') 'ERROR reading ',fieldname ; status = he5_swdetach(swid) 
      status = he5_swclose(fid) ; goto 100 
    end if
    status = he5_swrdlattr(swid,fieldname,"ScaleFactor",scalefactor)
    cloudFraction = real(int16_buffer) * scalefactor
    where(cloudfraction < (-32760.0*scalefactor)) cloudfraction = FILL_FLT32
  end if

  retstatus = retstatus +1 ; fieldname =  'CloudPressure'
  if (index(trim(dfieldlist),trim(fieldname)) > 0) then
    status = he5_swrdfld( swid, fieldname, start, stride, edge, int16_buffer)
    if (status /= 0 ) then 
      write(*,'(2a)') 'ERROR reading ',fieldname ; status = he5_swdetach(swid) 
      status = he5_swclose(fid) ; goto 100 
    end if
    cloudpressure=real(int16_buffer)
    where(cloudpressure < -32760.0) cloudpressure = FILL_FLT32
  end if

  retstatus = retstatus +1 ; fieldname =  'TerrainHeight'
  if (index(trim(dfieldlist),trim(fieldname)) > 0) then
    status = he5_swrdfld( swid, fieldname, start, stride, edge, terrainheight)
    if (status /= 0 ) then 
      write(*,'(2a)') 'ERROR reading ',fieldname ; status = he5_swdetach(swid) 
      status = he5_swclose(fid) ; goto 100 
    end if
  end if

  retstatus = retstatus +1 ; fieldname =  'TM4SurfacePressure'
  if (index(trim(dfieldlist),trim(fieldname)) > 0) then  
    status = he5_swrdfld( swid, fieldname, start, stride, edge, tm4surfacepressure)
    if (status /= 0 ) then 
      write(*,'(2a)') 'ERROR reading ',fieldname ; status = he5_swdetach(swid) 
      status = he5_swclose(fid) ; goto 100 
    end if
  end if

  retstatus = retstatus +1 ; fieldname =  'GhostColumn'
  if (index(trim(dfieldlist),trim(fieldname)) > 0) then
    status = he5_swrdfld( swid, fieldname, start, stride, edge, GhostColumn)
    if (status /= 0 ) then 
      write(*,'(2a)') 'ERROR reading ',fieldname ; status = he5_swdetach(swid) 
      status = he5_swclose(fid) ; goto 100 
    end if
  end if

  retstatus = retstatus +1 ; fieldname =  'AirMassFactor'
  if (index(trim(dfieldlist),trim(fieldname)) > 0) then
    status = he5_swrdfld( swid, fieldname, start, stride, edge,AirMassFactor )
    if (status /= 0 ) then 
      write(*,'(2a)') 'ERROR reading ',fieldname ; status = he5_swdetach(swid) 
      status = he5_swclose(fid) ; goto 100 
    end if
  end if

  retstatus = retstatus +1 ; fieldname =  'AirMassFactorTropospheric'
  if (index(trim(dfieldlist),trim(fieldname)) > 0) then
    status = he5_swrdfld( swid, fieldname, start, stride, edge, AirMassFactorTropospheric)
    if (status /= 0 ) then 
      write(*,'(2a)') 'ERROR reading ',fieldname ; status = he5_swdetach(swid) 
      status = he5_swclose(fid) ; goto 100 
    end if
  end if
  
  retstatus = retstatus +1 ; fieldname =  'AirMassFactorGeometric'
  if (index(trim(dfieldlist),trim(fieldname)) > 0) then
    status = he5_swrdfld( swid, fieldname, start, stride, edge, AirMassFactorGeometric)
    if (status /= 0 ) then 
      write(*,'(2a)') 'ERROR reading ',fieldname ; status = he5_swdetach(swid) 
      status = he5_swclose(fid) ; goto 100 
    end if
  end if

  retstatus = retstatus +1 ; fieldname =  'SurfaceAlbedo'
  if (index(trim(dfieldlist),trim(fieldname)) > 0) then
    status = he5_swrdfld( swid, fieldname, start, stride, edge, int16_buffer)
    if (status /= 0 ) then 
      write(*,'(2a)') 'ERROR reading ',fieldname ; status = he5_swdetach(swid) 
      status = he5_swclose(fid) ; goto 100 
    end if
    status = he5_swrdlattr(swid,fieldname,"ScaleFactor",scalefactor)
    surfacealbedo = real(int16_buffer) * scalefactor
    where(surfacealbedo < (-32760.0*scalefactor)) surfacealbedo = FILL_FLT32
  end if

  retstatus = retstatus +1 ; fieldname =  'VCDTropErrorUsingAvKernel'
  if (index(trim(dfieldlist),trim(fieldname)) > 0) then
    status = he5_swrdfld( swid, fieldname, start, stride, edge, VCDTropErrorUsingAvKernel)
    if (status /= 0 ) then 
      write(*,'(2a)') 'ERROR reading ',fieldname ; status = he5_swdetach(swid) 
      status = he5_swclose(fid) ; goto 100 
    end if
  end if

  retstatus = retstatus +1 ; fieldname =  'VCDErrorUsingAvKernel'
  if (index(trim(dfieldlist),trim(fieldname)) > 0) then
    status = he5_swrdfld( swid, fieldname, start, stride, edge, VCDErrorUsingAvKernel)
    if (status /= 0 ) then 
      write(*,'(2a)') 'ERROR reading ',fieldname ; status = he5_swdetach(swid) 
      status = he5_swclose(fid) ; goto 100 
    end if
  end if

  retstatus = retstatus +1 ; fieldname =  'CloudRadianceFraction'
  if (index(trim(dfieldlist),trim(fieldname)) > 0) then
    status = he5_swrdfld( swid, fieldname, start, stride, edge, int16_buffer)
    if (status /= 0 ) then 
      write(*,'(2a)') 'ERROR reading ',fieldname ; status = he5_swdetach(swid) 
      status = he5_swclose(fid) ; goto 100 
    end if
    status = he5_swrdlattr(swid,fieldname,"ScaleFactor",scalefactor)
    CloudRadianceFraction = real(int16_buffer) * scalefactor    
    where(CloudRadianceFraction < (-32760.0 * scalefactor ) ) CloudRadianceFraction = FILL_FLT32
  end if

  retstatus = retstatus +1 ; fieldname =  'TroposphericVerticalColumnError'
  if (index(trim(dfieldlist),trim(fieldname)) > 0) then
    status = he5_swrdfld( swid, fieldname, start, stride, edge, TroposphericVerticalColumnError)
    if (status /= 0 ) then 
      write(*,'(2a)') 'ERROR reading ',fieldname ; status = he5_swdetach(swid) 
      status = he5_swclose(fid) ; goto 100 
    end if
  end if

  retstatus = retstatus +1 ; fieldname =  'TotalVerticalColumn'
  if (index(trim(dfieldlist),trim(fieldname)) > 0) then
    status = he5_swrdfld( swid, fieldname, start, stride, edge,TotalVerticalColumn )
    if (status /= 0 ) then 
      write(*,'(2a)') 'ERROR reading ',fieldname ; status = he5_swdetach(swid) 
      status = he5_swclose(fid) ; goto 100 
    end if
  end if

  retstatus = retstatus +1 ; fieldname =  'TotalVerticalColumnError'
  if (index(trim(dfieldlist),trim(fieldname)) > 0) then
    status = he5_swrdfld( swid, fieldname, start, stride, edge,TotalVerticalColumnError )
    if (status /= 0 ) then 
      write(*,'(2a)') 'ERROR reading ',fieldname ; status = he5_swdetach(swid) 
      status = he5_swclose(fid) ; goto 100 
    end if
  end if

  retstatus = retstatus +1 ; fieldname =  'TroposphericVerticalColumn'
  if (index(trim(dfieldlist),trim(fieldname)) > 0) then
    status = he5_swrdfld( swid, fieldname, start, stride, edge, TroposphericVerticalColumn)
    if (status /= 0 ) then 
      write(*,'(2a)') 'ERROR reading ',fieldname ; status = he5_swdetach(swid) 
      status = he5_swclose(fid) ; goto 100 
    end if
  end if

  retstatus = retstatus +1 ; fieldname =  'AssimilatedStratosphericSlantColumn'
  if (index(trim(dfieldlist),trim(fieldname)) > 0) then
    status = he5_swrdfld( swid, fieldname, start, stride, edge, AssStratSlantColumn)
    if (status /= 0 ) then 
      write(*,'(2a)') 'ERROR reading ',fieldname ; status = he5_swdetach(swid) 
      status = he5_swclose(fid) ; goto 100 
    end if
  end if

  retstatus = retstatus +1 ; fieldname =  'AssimilatedStratosphericVerticalColumn'
  if (index(trim(dfieldlist),trim(fieldname)) > 0) then
    status = he5_swrdfld( swid, fieldname, start, stride, edge,AssStratVerticalColumn )
    if (status /= 0 ) then 
      write(*,'(2a)') 'ERROR reading ',fieldname ; status = he5_swdetach(swid) 
      status = he5_swclose(fid) ; goto 100 
    end if
  end if

  retstatus = retstatus +1 ; fieldname =  'AssimilatedStratosphericVerticalColumnError'
  if (index(trim(dfieldlist),trim(fieldname)) > 0) then
    status = he5_swrdfld( swid, fieldname, start, stride, edge, AssStratVerticalColumnError)
    if (status /= 0 ) then 
      write(*,'(2a)') 'ERROR reading ',fieldname ; status = he5_swdetach(swid) 
      status = he5_swclose(fid) ; goto 100 
    end if
  end if

  retstatus = retstatus +1 ; fieldname =  'SlantColumnAmountNO2'
  if (index(trim(dfieldlist),trim(fieldname)) > 0) then
    status = he5_swrdfld( swid, fieldname, start, stride, edge, SlantColumnAmountNO2)
    if (status /= 0 ) then 
      write(*,'(2a)') 'ERROR reading ',fieldname ; status = he5_swdetach(swid) 
      status = he5_swclose(fid) ; goto 100 
    end if
  end if

  retstatus = retstatus +1 ; fieldname =  'SlantColumnAmountNO2Std'
  if (index(trim(dfieldlist),trim(fieldname)) > 0) then
    status = he5_swrdfld( swid, fieldname, start, stride, edge, SlantColumnAmountNO2Std)
    if (status /= 0 ) then 
      write(*,'(2a)') 'ERROR reading ',fieldname ; status = he5_swdetach(swid) 
      status = he5_swclose(fid) ; goto 100 
    end if
  end if

  retstatus = retstatus +1 ; fieldname =  'TroposphericVerticalColumnModel'
  if (index(trim(dfieldlist),trim(fieldname)) > 0) then
    status = he5_swrdfld( swid, fieldname, start, stride, edge,TroposphericVerticalColumnModel )
    if (status /= 0 ) then 
      write(*,'(2a)') 'ERROR reading ',fieldname ; status = he5_swdetach(swid) 
      status = he5_swclose(fid) ; goto 100 
    end if
  end if

  retstatus = retstatus +1 ; fieldname =  'TM4TropoPauseLevel'
  if (index(trim(dfieldlist),trim(fieldname)) > 0) then
    status = he5_swrdfld( swid, fieldname, start, stride, edge, TM4TropoPauseLevel)
    if (status /= 0 ) then 
      write(*,'(2a)') 'ERROR reading ',fieldname ; status = he5_swdetach(swid) 
      status = he5_swclose(fid) ; goto 100 
    end if
  end if

  retstatus = retstatus +1 ; fieldname =  'TM4TerrainHeight'
  if (index(trim(dfieldlist),trim(fieldname)) > 0) then
    status = he5_swrdfld( swid, fieldname, start, stride, edge, TM4TerrainHeight)
    if (status /= 0 ) then 
      write(*,'(2a)') 'ERROR reading ',fieldname ; status = he5_swdetach(swid) 
      status = he5_swclose(fid) ; goto 100 
    end if
  end if

  retstatus = retstatus +1 ; fieldname =  'TroposphericColumnFlag'
  if (index(trim(dfieldlist),trim(fieldname)) > 0) then
    status = he5_swrdfld( swid, fieldname, start, stride, edge, TroposphericColumnFlag)
    if (status /= 0 ) then 
      write(*,'(2a)') 'ERROR reading ',fieldname ; status = he5_swdetach(swid) 
      status = he5_swclose(fid) ; goto 100 
    end if
  end if

  retstatus = retstatus +1 ; fieldname =  'GroundPixelQualityFlags'
  if (index(trim(gfieldlist),trim(fieldname)) > 0) then
    status = he5_swrdfld( swid, fieldname, start, stride, edge, GroundPixelQualityFlags)
    if (status /= 0 ) then 
      write(*,'(2a)') 'ERROR reading ',fieldname ; status = he5_swdetach(swid) 
      status = he5_swclose(fid) ; goto 100 
    end if
  end if

  retstatus = retstatus +1 ; fieldname =  'CloudPressureStd'
  if (index(trim(dfieldlist),trim(fieldname)) > 0) then
    status = he5_swrdfld( swid, fieldname, start, stride, edge, int16_buffer)
    if (status /= 0 ) then 
      write(*,'(2a)') 'ERROR reading ',fieldname ; status = he5_swdetach(swid) 
      status = he5_swclose(fid) ; goto 100 
    end if
    CloudPressureStd=real(int16_buffer)
    where(CloudPressureStd < -32760.0)  CloudPressureStd = FILL_FLT32
  end if

  retstatus = retstatus +1 ; fieldname =  'CloudFractionStd'
  if (index(trim(dfieldlist),trim(fieldname)) > 0) then
    status = he5_swrdfld( swid, fieldname, start, stride, edge, int16_buffer)
    if (status /= 0 ) then 
      write(*,'(2a)') 'ERROR reading ',fieldname ; status = he5_swdetach(swid) 
      status = he5_swclose(fid) ; goto 100 
    end if
    status = he5_swrdlattr(swid,fieldname,"ScaleFactor",scalefactor)
    CloudFractionStd = real(int16_buffer) * scalefactor
    where(cloudfractionstd < (-32760.0*scalefactor)) cloudfractionstd = FILL_FLT32
  end if

  ! 3-D fields
  start(:)  = 0 ; stride(:) = 1 
  edge(1)   = nxtrack ; edge(2) = ntimes ; edge(3) = NCORNERPOINTS

  retstatus = retstatus +1 ; fieldname =  'LatitudeCornerpoints'
  if (index(trim(gfieldlist),trim(fieldname)) > 0) then
    status = he5_swrdfld( swid, fieldname, start, stride, edge, LatitudeCornerpoints)
    if (status /= 0 ) then 
      write(*,'(2a)') 'ERROR reading ',fieldname ; status = he5_swdetach(swid) 
      status = he5_swclose(fid) ; goto 100 
    end if
  end if

  retstatus = retstatus +1 ; fieldname =  'LongitudeCornerpoints'
  if (index(trim(gfieldlist),trim(fieldname)) > 0) then
    status = he5_swrdfld( swid, fieldname, start, stride, edge,LongitudeCornerpoints )
    if (status /= 0 ) then 
      write(*,'(2a)') 'ERROR reading ',fieldname ; status = he5_swdetach(swid) 
      status = he5_swclose(fid) ; goto 100 
    end if
  end if

  retstatus = retstatus +1 ; fieldname =  'AveragingKernel' ; edge(3)=npressurelevels
  if (index(trim(dfieldlist),trim(fieldname)) > 0) then
    status = he5_swrdfld( swid, fieldname, start, stride, edge,int16_buffer3d)
    if (status /= 0 ) then 
      write(*,'(2a)') 'ERROR reading ',fieldname ; status = he5_swdetach(swid) 
      status = he5_swclose(fid) ; goto 100 
    end if
    status = he5_swrdlattr(swid,fieldname,"ScaleFactor",scalefactor)
    averagingkernel = real(int16_buffer3d) * scalefactor
    where(averagingkernel < (-32760.0*scalefactor)) averagingkernel = FILL_FLT32
  end if

  ! Pressure levels
  retstatus = retstatus +1 ; fieldname =  'TM4PressurelevelA' ; edge(1)=npressurelevels
  if (index(trim(dfieldlist),trim(fieldname)) > 0) then
    status = he5_swrdfld( swid, fieldname, start, stride, edge,floatbuffer)
    if (status /= 0 ) then 
      write(*,'(2a)') 'ERROR reading ',fieldname ; status = he5_swdetach(swid) 
      status = he5_swclose(fid) ; goto 100 
    end if
    a_lev(1:npressurelevels) = floatbuffer
  end if

  retstatus = retstatus +1 ; fieldname =  'TM4PressurelevelB' 
  if (index(trim(dfieldlist),trim(fieldname)) > 0) then
    status = he5_swrdfld( swid, fieldname, start, stride, edge,floatbuffer)
    if (status /= 0 ) then 
      write(*,'(2a)') 'ERROR reading ',fieldname ; status = he5_swdetach(swid) 
      status = he5_swclose(fid) ; goto 100 
    end if
    b_lev(1:npressurelevels) = floatbuffer
  end if
  
  ! detach the swath & close the file
  status = he5_swdetach(swid) ; status = he5_swclose(fid)   

  if (kind(TM4TerrainHeight) /= kind(datastruct%TM4TerrainHeight) )                  stop 'datatype inconsistency for TM4TerrainHeight'

  datastruct(1:nxtrack,1:ntimes)%SolarZenithAngle    = SolarZenithAngle 
  datastruct(1:nxtrack,1:ntimes)%SolarAzimuthAngle   = SolarAzimuthAngle      
  datastruct(1:nxtrack,1:ntimes)%longitude           = longitude     
  datastruct(1:nxtrack,1:ntimes)%latitude            = latitude      
  datastruct(1:nxtrack,1:ntimes)%ViewingZenithAngle  = ViewingZenithAngle         
  datastruct(1:nxtrack,1:ntimes)%ViewingAzimuthAngle = ViewingAzimuthAngle       
  datastruct(1:nxtrack,1:ntimes)%CloudFraction       = CloudFraction
  datastruct(1:nxtrack,1:ntimes)%CloudPressure       = CloudPressure 
  datastruct(1:nxtrack,1:ntimes)%Terrainheight       = Terrainheight
  datastruct(1:nxtrack,1:ntimes)%tm4surfacepressure  = tm4surfacepressure  

  datastruct(1:nxtrack,1:ntimes)%GhostColumn                                  = GhostColumn                               
  datastruct(1:nxtrack,1:ntimes)%AirMassFactor                                = AirMassFactor                             
  datastruct(1:nxtrack,1:ntimes)%AirMassFactorTropospheric                    = AirMassFactorTropospheric                 
  datastruct(1:nxtrack,1:ntimes)%AirMassFactorGeometric                       = AirMassFactorGeometric                    
  datastruct(1:nxtrack,1:ntimes)%SurfaceAlbedo                                = SurfaceAlbedo                             
  datastruct(1:nxtrack,1:ntimes)%VCDTropErrorUsingAvKernel                    = VCDTropErrorUsingAvKernel                 
  datastruct(1:nxtrack,1:ntimes)%VCDErrorUsingAvKernel                        = VCDErrorUsingAvKernel                     
  datastruct(1:nxtrack,1:ntimes)%CloudRadianceFraction                        = CloudRadianceFraction                     
  datastruct(1:nxtrack,1:ntimes)%TroposphericVerticalColumnError              = TroposphericVerticalColumnError           
  datastruct(1:nxtrack,1:ntimes)%TotalVerticalColumn                          = TotalVerticalColumn                       
  datastruct(1:nxtrack,1:ntimes)%TotalVerticalColumnError                     = TotalVerticalColumnError                  
  datastruct(1:nxtrack,1:ntimes)%TroposphericVerticalColumn                   = TroposphericVerticalColumn                
  datastruct(1:nxtrack,1:ntimes)%AssStratSlantColumn                          = AssStratSlantColumn       
  datastruct(1:nxtrack,1:ntimes)%AssStratVerticalColumn                       = AssStratVerticalColumn    
  datastruct(1:nxtrack,1:ntimes)%AssStratVerticalColumnError                  = AssStratVerticalColumnError
  datastruct(1:nxtrack,1:ntimes)%SlantColumnAmountNO2                         = SlantColumnAmountNO2                      
  datastruct(1:nxtrack,1:ntimes)%SlantColumnAmountNO2Std                      = SlantColumnAmountNO2Std                   
  datastruct(1:nxtrack,1:ntimes)%TroposphericVerticalColumnModel              = TroposphericVerticalColumnModel           
  datastruct(1:nxtrack,1:ntimes)%TM4TropoPauseLevel                           = TM4TropoPauseLevel
  datastruct(1:nxtrack,1:ntimes)%TM4TerrainHeight                             = TM4TerrainHeight
  datastruct(1:nxtrack,1:ntimes)%TroposphericColumnFlag                       = TroposphericColumnFlag
  datastruct(1:nxtrack,1:ntimes)%GroundPixelQualityFlags                      = GroundPixelQualityFlags
  datastruct(1:nxtrack,1:ntimes)%CloudPressureStd                             = CloudPressureStd 
  datastruct(1:nxtrack,1:ntimes)%CloudFractionStd                             = CloudFractionStd
  do i=1,NCORNERPOINTS
    datastruct(1:nxtrack,1:ntimes)%LatitudeCornerpoints(i)                    = LatitudeCornerpoints(:,:,i)
    datastruct(1:nxtrack,1:ntimes)%LongitudeCornerpoints(i)                   = LongitudeCornerpoints(:,:,i)   
  end do
  do i=1,npressurelevels
    datastruct(1:nxtrack,1:ntimes)%averagingkernel(i)                         = averagingkernel(:,:,i)
  end do

  
  datastruct(1,1:ntimes)%TAItime                     = TAItime  
  datastruct(1,1:ntimes)%htime                       = htime
  datastruct(1,1:ntimes)%hdate                       = hdate
  datastruct(1,1:ntimes)%MeasurementQualityFlags     = MeasurementQualityFlags
  datastruct(1,1:ntimes)%InstrumentConfigurationId   = InstrumentConfigurationId
  datastruct(1,1:ntimes)%SpacecraftAltitude          = SpacecraftAltitude
  datastruct(1,1:ntimes)%SpacecraftLatitude          = SpacecraftLatitude
  datastruct(1,1:ntimes)%SpacecraftLongitude         = SpacecraftLongitude
      
  do i = 2,nxtrack
    datastruct(i,:)%TAItime                   = datastruct(1,:)%TAItime
    datastruct(i,:)%htime                     = datastruct(1,:)%htime
    datastruct(i,:)%hdate                     = datastruct(1,:)%hdate
    datastruct(i,:)%MeasurementQualityFlags   = datastruct(1,:)%MeasurementQualityFlags
    datastruct(i,:)%InstrumentConfigurationId = datastruct(1,:)%InstrumentConfigurationId    
    datastruct(i,:)%SpacecraftLatitude        = datastruct(1,:)%SpacecraftLatitude
    datastruct(i,:)%SpacecraftLongitude       = datastruct(1,:)%SpacecraftLongitude
    datastruct(i,:)%SpacecraftAltitude        = datastruct(1,:)%SpacecraftAltitude
  end do
  
  where(datastruct%GroundPixelQualityFlags < 0)   datastruct%GroundPixelQualityFlags   = datastruct%GroundPixelQualityFlags   + 65536
  where(datastruct%MeasurementQualityFlags < 0)   datastruct%MeasurementQualityFlags   = datastruct%MeasurementQualityFlags   + 256
  where(datastruct%InstrumentConfigurationId < 0) datastruct%InstrumentConfigurationId = datastruct%InstrumentConfigurationId + 256
  where(datastruct%TM4TropoPauseLevel <0)         datastruct%TM4TropoPauseLevel        = datastruct%TM4TropoPauseLevel        + 256
  
! everything succesful
  retStatus = 0

  !  deallocate temporary arrays
100 if (allocated(latitude))                                   deallocate(latitude)                                   
  if (allocated(longitude))                                    deallocate(longitude)                                  
  if (allocated(SolarZenithAngle))                             deallocate(SolarZenithAngle)                           
  if (allocated(SolarAzimuthAngle))                            deallocate(SolarAzimuthAngle)                          
  if (allocated(ViewingZenithAngle))                           deallocate(ViewingZenithAngle)                         
  if (allocated(ViewingAzimuthAngle))                          deallocate(ViewingAzimuthAngle)                        
  if (allocated(cloudfraction))                                deallocate(cloudfraction)                              
  if (allocated(cloudpressure))                                deallocate(cloudpressure)                              
  if (allocated(terrainheight))                                deallocate(terrainheight)                              
  if (allocated(TM4SurfacePressure))                           deallocate(TM4SurfacePressure)                         
  if (allocated(TAItime))                                      deallocate(TAItime)
  if (allocated(htime))                                        deallocate(htime)
  if (allocated(hdate))                                        deallocate(hdate)
  if (allocated(SpacecraftAltitude))                           deallocate(SpacecraftAltitude)                         
  if (allocated(SpacecraftLatitude))                           deallocate(SpacecraftLatitude)                         
  if (allocated(SpacecraftLongitude))                          deallocate(SpacecraftLongitude)                        
  if (allocated(GhostColumn))                                  deallocate(GhostColumn)                                
  if (allocated(AirMassFactor))                                deallocate(AirMassFactor)                              
  if (allocated(AirMassFactorTropospheric))                    deallocate(AirMassFactorTropospheric)                  
  if (allocated(AirMassFactorGeometric))                       deallocate(AirMassFactorGeometric)                     
  if (allocated(SurfaceAlbedo))                                deallocate(SurfaceAlbedo)                              
  if (allocated(VCDTropErrorUsingAvKernel))                    deallocate(VCDTropErrorUsingAvKernel)                  
  if (allocated(VCDErrorUsingAvKernel))                        deallocate(VCDErrorUsingAvKernel)                      
  if (allocated(CloudRadianceFraction))                        deallocate(CloudRadianceFraction)                      
  if (allocated(TroposphericVerticalColumnError))              deallocate(TroposphericVerticalColumnError)            
  if (allocated(TotalVerticalColumn))                          deallocate(TotalVerticalColumn)                        
  if (allocated(TotalVerticalColumnError))                     deallocate(TotalVerticalColumnError)                   
  if (allocated(TroposphericVerticalColumn))                   deallocate(TroposphericVerticalColumn)                 
  if (allocated(AssStratSlantColumn))                          deallocate(AssStratSlantColumn)        
  if (allocated(AssStratVerticalColumn))                       deallocate(AssStratVerticalColumn)     
  if (allocated(AssStratVerticalColumnError))                  deallocate(AssStratVerticalColumnError)
  if (allocated(SlantColumnAmountNO2))                         deallocate(SlantColumnAmountNO2)                       
  if (allocated(SlantColumnAmountNO2Std))                      deallocate(SlantColumnAmountNO2Std)                    
  if (allocated(TroposphericVerticalColumnModel))              deallocate(TroposphericVerticalColumnModel)            
  if (allocated(LatitudeCornerpoints))                         deallocate(LatitudeCornerpoints)                       
  if (allocated(LongitudeCornerpoints))                        deallocate(LongitudeCornerpoints)                      
  if (allocated(TM4TropoPauseLevel))                           deallocate(TM4TropoPauseLevel)                         
  if (allocated(TM4TerrainHeight))                             deallocate(TM4TerrainHeight)                           
  if (allocated(TroposphericColumnFlag))                       deallocate(TroposphericColumnFlag)                     
  if (allocated(InstrumentConfigurationId))                    deallocate(InstrumentConfigurationId)           
  if (allocated(MeasurementQualityFlags))                      deallocate(MeasurementQualityFlags)             
  if (allocated(GroundPixelQualityFlags))                      deallocate(GroundPixelQualityFlags)                    
  if (allocated(CloudPressureStd))                             deallocate(CloudPressureStd)
  if (allocated(CloudFractionStd))                             deallocate(CloudFractionStd)
  if (allocated(int16_buffer))                                 deallocate(int16_buffer)
  if (allocated(int32_buffer))                                 deallocate(int32_buffer)
  if (allocated(int8_buffer))                                  deallocate(int8_buffer)
  if (allocated(averagingkernel))                              deallocate(averagingkernel)
  if (allocated(int16_buffer3d))                               deallocate(int16_buffer3d)
  if (allocated(floatbuffer))                                  deallocate(floatbuffer)
  
end subroutine read_OMI_tropNO2                               

!**************************************

subroutine read_omi_tropno2_metadata(filename,metadata,retstatus)
  use metadata_module

  implicit none
  character(len=STD_STRING),intent(in)        :: filename
  character(len=metadata_value_length)        :: strbuf
  character(len=SWATHNAME_LENGTH)             :: swathname,attrname,dumstr
  character(len=3000)                         :: attrnamelist
  character(len=SWATHNAME_LENGTH),dimension(:),allocatable :: attrnames
  integer                                     :: i,nattributes,listlength,numbuf
  integer                                     :: swid,fid,status,retstatus,nt,cnt,N
  integer                                     :: inputpointer
  logical                                     :: fexist
  type(Tmetadata),dimension(NMetadataFields)  :: metadata
  
  
  N=NMetadataFields
  retstatus = 1 ; fid = -1 ; swid = -1
  write(swathname,'(a)') '                                                          '
  write(swathname,'(a)') 'DominoNO2'
  inquire (file=filename,exist=fexist)
  
  if (fexist == .false.) then
    write(*,'(3a)') 'read_omi_tropno2_metadata: ERROR file ',trim(filename),' does not exist'
    return
  end if

  ! open the file with read-only access
  fid = he5_swopen( filename, HE5F_ACC_RDONLY )         
  if (fid <= 0 ) then
    write(*,'(2a)') 'read_omi_tropno2_metadata: ERROR cannot open file ',trim(filename)
    return
  end if
  
  ! attach to the swath
  swid = he5_swattach( fid, swathname )  
  if (swid <= 0 ) then
    write(*,'(2a)') 'read_omi_tropno2_metadata: ERROR cannot attach swath ',trim(swathname)
    status = he5_swclose(fid)
    return
  end if
  
  nattributes = he5_swinqattrs(swid,attrnamelist,listlength)
  allocate(attrnames(nattributes))
  call separate_attributenames(attrnamelist,attrnames,nattributes)

  if ( nattributes < 1) then
    write(*,'(a)') 'There are no attributes in this file'
    return
  end if

  if (nattributes > N) then
    write(*,'(a,i3,a,i3,a)') 'read_omi_tropno2_metadata: # attributes exceeds allocated memory space. Only ',N,' of ',nattributes,' attributes are read'
    nattributes=N
  end if

  do i=1,N
    write(metadata(i)%value,'(a)') '                                           '
    write(metadata(i)%fieldname,'(a)') '                                           '    
    metadata(i)%used=.false.
  end do

  inputpointer=1
  do i=1,nattributes
    write(strbuf,'(a)') '                                          '
    write(attrname,'(a)') '                                        '
    write(attrname,'(a)') trim(attrnames(i))
    write(metadata(i)%fieldname,'(a)') trim(attrname)
    if ((index(trim(attrname),'METEO_DATA') > 0) .OR. &
        (index(trim(attrname),'NO2_L2') > 0)     .OR. &
        (index(trim(attrname),'AMF_LUT') > 0)) then
      write(dumstr,'(i2)') inputpointer
      write(metadata(i)%fieldname,'(a)') '                                       '
      write(metadata(i)%fieldname,'(2a)') 'Input_pointer_',trim(adjustl(dumstr))
      inputpointer=inputpointer+1
    end if
    status = he5_swattrinfo(swid,trim(attrname),nt,cnt)
    metadata(i)%numvalue=-999    
    write(metadata(i)%value,'(a)') 'N/A'    
    if (nt == 57) then
      status = he5_swrdattr(swid, trim(attrname), strbuf)  
      if (status /= 0 ) then
        write(*,'(2a)') 'Error reading swath attribute ',trim(attrname)
        cycle
      end if
      write(metadata(i)%value,'(a)') trim(strbuf)      
    else
      write(metadata(i)%value,'(a)') '                                 '
      status = he5_swrdattr(swid, trim(attrname), numbuf)
      if (status /= 0 ) then
        write(*,'(2a)') 'Error reading swath attribute ',trim(attrname)
        cycle
      end if
      metadata(i)%numvalue=numbuf
    end if      
    metadata(i)%used=.true.
  end do

  ! detach the swath & close the file
  status = he5_swdetach(swid) ; status = he5_swclose(fid)   
  
  retstatus = 0
  if (allocated(attrnames)) deallocate(attrnames)
  
end subroutine read_omi_tropno2_metadata

!*****************************************************

subroutine separate_attributenames(attrnamelist,attrnames,nattributes)
  implicit none
  
  integer,intent(in)              :: nattributes
  character(len=SWATHNAME_LENGTH) :: attrnames(nattributes)
  character(len=3000)             :: attrnamelist,dumlist
  integer                         :: i,pos,ii
  
  ii=0
  do i=1,nattributes-1
    write(attrnames(i),'(a)') '                                 '
    pos=index(attrnamelist,',')
    if (pos < 1) cycle
    ii=ii+1
    write(attrnames(ii),'(a)') trim(attrnamelist(1:pos-1))
    write(dumlist,'(a)') trim(attrnamelist(pos+1:len(trim(attrnamelist))))
    write(attrnamelist,'(a)') trim(dumlist)
  end do
  if (len(trim(attrnamelist)) > 1 ) write(attrnames(ii+1),'(a)') trim(attrnamelist)

end subroutine separate_attributenames

!*****************************************************

subroutine copy_data2D(data_in,data_out,nxtrack,ntimes,npressurelevels)
  implicit none
  integer,intent(in)                                             :: nxtrack,ntimes,npressurelevels
  integer                                                        :: i
  type(tdatastruct),dimension(max_nxtrack,max_ntimes),intent(in) :: data_in
  type(tdatastruct),dimension(nxtrack,ntimes)                    :: data_out


  data_out%SolarZenithAngle                            = data_in(1:nxtrack,1:ntimes)%SolarZenithAngle
  data_out%SolarAzimuthAngle                           = data_in(1:nxtrack,1:ntimes)%SolarAzimuthAngle
  data_out%longitude                                   = data_in(1:nxtrack,1:ntimes)%longitude
  data_out%latitude                                    = data_in(1:nxtrack,1:ntimes)%latitude
  data_out%ViewingZenithAngle                          = data_in(1:nxtrack,1:ntimes)%ViewingZenithAngle
  data_out%ViewingAzimuthAngle                         = data_in(1:nxtrack,1:ntimes)%ViewingAzimuthAngle
  data_out%TM4SurfacePressure                          = data_in(1:nxtrack,1:ntimes)%TM4SurfacePressure
  data_out%CloudPressure                               = data_in(1:nxtrack,1:ntimes)%CloudPressure
  data_out%CloudPressureStd                            = data_in(1:nxtrack,1:ntimes)%CloudPressureStd
  data_out%CloudFraction                               = data_in(1:nxtrack,1:ntimes)%CloudFraction
  data_out%CloudFractionStd                            = data_in(1:nxtrack,1:ntimes)%CloudFractionStd
  data_out%SpacecraftAltitude                          = data_in(1:nxtrack,1:ntimes)%SpacecraftAltitude
  data_out%SpacecraftLatitude                          = data_in(1:nxtrack,1:ntimes)%SpacecraftLatitude
  data_out%SpacecraftLongitude                         = data_in(1:nxtrack,1:ntimes)%SpacecraftLongitude
  data_out%GhostColumn                                 = data_in(1:nxtrack,1:ntimes)%GhostColumn
  data_out%AirMassFactor                               = data_in(1:nxtrack,1:ntimes)%AirMassFactor
  data_out%AirMassFactorTropospheric                   = data_in(1:nxtrack,1:ntimes)%AirMassFactorTropospheric
  data_out%AirMassFactorGeometric                      = data_in(1:nxtrack,1:ntimes)%AirMassFactorGeometric
  data_out%SurfaceAlbedo                               = data_in(1:nxtrack,1:ntimes)%SurfaceAlbedo
  data_out%VCDTropErrorUsingAvKernel                   = data_in(1:nxtrack,1:ntimes)%VCDTropErrorUsingAvKernel
  data_out%VCDErrorUsingAvKernel                       = data_in(1:nxtrack,1:ntimes)%VCDErrorUsingAvKernel
  data_out%CloudRadianceFraction                       = data_in(1:nxtrack,1:ntimes)%CloudRadianceFraction
  data_out%TotalVerticalColumn                         = data_in(1:nxtrack,1:ntimes)%TotalVerticalColumn
  data_out%TotalVerticalColumnError                    = data_in(1:nxtrack,1:ntimes)%TotalVerticalColumnError
  data_out%TroposphericVerticalColumn                  = data_in(1:nxtrack,1:ntimes)%TroposphericVerticalColumn
  data_out%TroposphericVerticalColumnError             = data_in(1:nxtrack,1:ntimes)%TroposphericVerticalColumnError          
  data_out%AssStratSlantColumn                         = data_in(1:nxtrack,1:ntimes)%AssStratSlantColumn
  data_out%AssStratVerticalColumn                      = data_in(1:nxtrack,1:ntimes)%AssStratVerticalColumn
  data_out%AssStratVerticalColumnError                 = data_in(1:nxtrack,1:ntimes)%AssStratVerticalColumnError
  data_out%SlantColumnAmountNO2                        = data_in(1:nxtrack,1:ntimes)%SlantColumnAmountNO2
  data_out%SlantColumnAmountNO2Std                     = data_in(1:nxtrack,1:ntimes)%SlantColumnAmountNO2Std
  data_out%TroposphericVerticalColumnModel             = data_in(1:nxtrack,1:ntimes)%TroposphericVerticalColumnModel
  data_out%TAItime                                     = data_in(1:nxtrack,1:ntimes)%TAItime
  data_out%htime                                       = data_in(1:nxtrack,1:ntimes)%htime
  data_out%hdate                                       = data_in(1:nxtrack,1:ntimes)%hdate
  data_out%MeasurementQualityFlags                     = data_in(1:nxtrack,1:ntimes)%MeasurementQualityFlags
  data_out%InstrumentConfigurationId                   = data_in(1:nxtrack,1:ntimes)%instrumentConfigurationId
  data_out%GroundPixelQualityFlags                     = data_in(1:nxtrack,1:ntimes)%GroundPixelQualityFlags
  data_out%TerrainHeight                               = data_in(1:nxtrack,1:ntimes)%TerrainHeight
  data_out%TM4TerrainHeight                            = data_in(1:nxtrack,1:ntimes)%TM4TerrainHeight
  data_out%TM4TropoPauseLevel                          = data_in(1:nxtrack,1:ntimes)%TM4TropoPauseLevel
  data_out%TroposphericColumnFlag                      = data_in(1:nxtrack,1:ntimes)%TroposphericColumnFlag

  do i=1,NCORNERPOINTS
    data_out%LatitudeCornerpoints(i)   = data_in(1:nxtrack,1:ntimes)%LatitudeCornerpoints(i)
    data_out%LongitudeCornerpoints(i)  = data_in(1:nxtrack,1:ntimes)%LongitudeCornerpoints(i)
  end do

  do i=1,npressurelevels
    data_out%averagingkernel(i)  = data_in(1:nxtrack,1:ntimes)%averagingkernel(i)
  end do

end subroutine copy_data2D

!************************************************************************

subroutine copy_data1D(data_in,data_out,index_array,nxtrack,ntimes,npressurelevels,N)
  implicit none
  integer,intent(in)                                             :: nxtrack,ntimes,npressurelevels,N
  integer                                                        :: i,j,ii,jj
  integer,dimension(nxtrack*ntimes)                              :: index_array
  type(tdatastruct),dimension(nxtrack,ntimes),intent(in)         :: data_in
  type(tdatastruct),dimension(N)                                 :: data_out

  do i=1,N
    ii = mod(index_array(i),nxtrack)+1 ! row number
    jj = index_array(i)/nxtrack      ! image number
    
    data_out(i)%imagenr                         = jj
    data_out(i)%pixelnr                         = ii
    data_out(i)%SolarZenithAngle                = data_in(ii,jj)%SolarZenithAngle
    data_out(i)%SolarAzimuthAngle               = data_in(ii,jj)%SolarAzimuthAngle
    data_out(i)%longitude                       = data_in(ii,jj)%longitude
    data_out(i)%latitude                        = data_in(ii,jj)%latitude
    data_out(i)%ViewingZenithAngle              = data_in(ii,jj)%ViewingZenithAngle
    data_out(i)%ViewingAzimuthAngle             = data_in(ii,jj)%ViewingAzimuthAngle
    data_out(i)%TM4SurfacePressure              = data_in(ii,jj)%TM4SurfacePressure
    data_out(i)%CloudPressure                   = data_in(ii,jj)%CloudPressure
    data_out(i)%CloudPressureStd                = data_in(ii,jj)%CloudPressureStd
    data_out(i)%CloudFraction                   = data_in(ii,jj)%CloudFraction
    data_out(i)%CloudFractionStd                = data_in(ii,jj)%CloudFractionStd
    data_out(i)%SpacecraftAltitude              = data_in(ii,jj)%SpacecraftAltitude
    data_out(i)%SpacecraftLatitude              = data_in(ii,jj)%SpacecraftLatitude
    data_out(i)%SpacecraftLongitude             = data_in(ii,jj)%SpacecraftLongitude
    data_out(i)%GhostColumn                     = data_in(ii,jj)%GhostColumn
    data_out(i)%AirMassFactor                   = data_in(ii,jj)%AirMassFactor
    data_out(i)%AirMassFactorTropospheric       = data_in(ii,jj)%AirMassFactorTropospheric
    data_out(i)%AirMassFactorGeometric          = data_in(ii,jj)%AirMassFactorGeometric
    data_out(i)%SurfaceAlbedo                   = data_in(ii,jj)%SurfaceAlbedo
    data_out(i)%VCDTropErrorUsingAvKernel       = data_in(ii,jj)%VCDTropErrorUsingAvKernel
    data_out(i)%VCDErrorUsingAvKernel           = data_in(ii,jj)%VCDErrorUsingAvKernel
    data_out(i)%CloudRadianceFraction           = data_in(ii,jj)%CloudRadianceFraction
    data_out(i)%TotalVerticalColumn             = data_in(ii,jj)%TotalVerticalColumn
    data_out(i)%TotalVerticalColumnError        = data_in(ii,jj)%TotalVerticalColumnError
    data_out(i)%TroposphericVerticalColumn      = data_in(ii,jj)%TroposphericVerticalColumn
    data_out(i)%TroposphericVerticalColumnError = data_in(ii,jj)%TroposphericVerticalColumnError          
    data_out(i)%AssStratSlantColumn             = data_in(ii,jj)%AssStratSlantColumn
    data_out(i)%AssStratVerticalColumn          = data_in(ii,jj)%AssStratVerticalColumn
    data_out(i)%AssStratVerticalColumnError     = data_in(ii,jj)%AssStratVerticalColumnError
    data_out(i)%SlantColumnAmountNO2            = data_in(ii,jj)%SlantColumnAmountNO2
    data_out(i)%SlantColumnAmountNO2Std         = data_in(ii,jj)%SlantColumnAmountNO2Std
    data_out(i)%TroposphericVerticalColumnModel = data_in(ii,jj)%TroposphericVerticalColumnModel
    data_out(i)%TAItime                         = data_in(ii,jj)%TAItime
    data_out(i)%htime                           = data_in(ii,jj)%htime
    data_out(i)%hdate                           = data_in(ii,jj)%hdate
    data_out(i)%MeasurementQualityFlags         = data_in(ii,jj)%MeasurementQualityFlags
    data_out(i)%InstrumentConfigurationId       = data_in(ii,jj)%InstrumentConfigurationId
    data_out(i)%GroundPixelQualityFlags         = data_in(ii,jj)%GroundPixelQualityFlags
    data_out(i)%TerrainHeight                   = data_in(ii,jj)%TerrainHeight
    data_out(i)%TM4TerrainHeight                = data_in(ii,jj)%TM4TerrainHeight
    data_out(i)%TM4TropoPauseLevel              = data_in(ii,jj)%TM4TropoPauseLevel
    data_out(i)%TroposphericColumnFlag          = data_in(ii,jj)%TroposphericColumnFlag
    do j=1,NCORNERPOINTS
      data_out(i)%LatitudeCornerpoints(j)       = data_in(ii,jj)%LatitudeCornerpoints(j)
      data_out(i)%LongitudeCornerpoints(j)      = data_in(ii,jj)%LongitudeCornerpoints(j)
    end do
    do j=1,npressurelevels ; data_out(i)%averagingkernel(j) = data_in(ii,jj)%averagingkernel(j) ; end do
  end do

end subroutine copy_data1D

end module read_hdfeos_module
